@@ -1,8 +1,9 @@
 import warnings
 from ssl import SSLContext
-from typing import AsyncIterator, Callable, Dict, List, Optional, Set, Tuple
+from typing import AsyncIterator, Callable, Dict, List, Optional, Set, Tuple, cast
 
-from .._backends.auto import AsyncLock, AsyncSemaphore, AutoBackend
+from .._backends.auto import AsyncLock, AsyncSemaphore
+from .._backends.base import lookup_async_backend
 from .._exceptions import LocalProtocolError, PoolTimeout, UnsupportedProtocol
 from .._threadlock import ThreadLock
 from .._types import URL, Headers, Origin, TimeoutDict
@@ -25,7 +26,7 @@ class NullSemaphore(AsyncSemaphore):
     async def acquire(self, timeout: float = None) -> None:
         return
 
-    def release(self) -> None:
+    async def release(self) -> None:
         return
 
 
@@ -37,7 +38,7 @@ class ResponseByteStream(AsyncByteStream):
         callback: Callable,
     ) -> None:
         """
-        A wrapper around the response stream that we return from `.request()`.
+        A wrapper around the response stream that we return from `.arequest()`.
 
         Ensures that when `stream.aclose()` is called, the connection pool
         is notified via a callback.
@@ -52,12 +53,12 @@ class ResponseByteStream(AsyncByteStream):
 
     async def aclose(self) -> None:
         try:
-            #  Call the underlying stream close callback.
+            # Call the underlying stream close callback.
             # This will be a call to `AsyncHTTP11Connection._response_closed()`
             # or `AsyncHTTP2Stream._response_closed()`.
             await self.stream.aclose()
         finally:
-            #  Call the connection pool close callback.
+            # Call the connection pool close callback.
             # This will be a call to `AsyncConnectionPool._response_closed()`.
             await self.callback(self.connection)
 
@@ -83,6 +84,7 @@ class AsyncConnectionPool(AsyncHTTPTransport):
     `local_address="0.0.0.0"` will connect using an `AF_INET` address (IPv4),
     while using `local_address="::"` will connect using an `AF_INET6` address
     (IPv6).
+    * **backend** - `str` - A name indicating which concurrency backend to use.
     """
 
     def __init__(
@@ -95,6 +97,7 @@ class AsyncConnectionPool(AsyncHTTPTransport):
         uds: str = None,
         local_address: str = None,
         max_keepalive: int = None,
+        backend: str = "auto",
     ):
         if max_keepalive is not None:
             warnings.warn(
@@ -112,7 +115,7 @@ class AsyncConnectionPool(AsyncHTTPTransport):
         self._local_address = local_address
         self._connections: Dict[Origin, Set[AsyncHTTPConnection]] = {}
         self._thread_lock = ThreadLock()
-        self._backend = AutoBackend()
+        self._backend = lookup_async_backend(backend)
         self._next_keepalive_check = 0.0
 
         if http2:
@@ -144,14 +147,27 @@ class AsyncConnectionPool(AsyncHTTPTransport):
             self._internal_connection_acquiry_lock = self._backend.create_lock()
         return self._internal_connection_acquiry_lock
 
-    async def request(
+    def _create_connection(
+        self,
+        origin: Tuple[bytes, bytes, int],
+    ) -> AsyncHTTPConnection:
+        return AsyncHTTPConnection(
+            origin=origin,
+            http2=self._http2,
+            uds=self._uds,
+            ssl_context=self._ssl_context,
+            local_address=self._local_address,
+            backend=self._backend,
+        )
+
+    async def arequest(
         self,
         method: bytes,
         url: URL,
         headers: Headers = None,
         stream: AsyncByteStream = None,
-        timeout: TimeoutDict = None,
-    ) -> Tuple[bytes, int, bytes, Headers, AsyncByteStream]:
+        ext: dict = None,
+    ) -> Tuple[int, Headers, AsyncByteStream, dict]:
         if url[0] not in (b"http", b"https"):
             scheme = url[0].decode("latin-1")
             raise UnsupportedProtocol(f"Unsupported URL protocol {scheme!r}")
@@ -159,6 +175,8 @@ class AsyncConnectionPool(AsyncHTTPTransport):
             raise LocalProtocolError("Missing hostname in URL.")
 
         origin = url_to_origin(url)
+        ext = {} if ext is None else ext
+        timeout = cast(TimeoutDict, ext.get("timeout", {}))
 
         await self._keepalive_sweep()
 
@@ -172,33 +190,28 @@ class AsyncConnectionPool(AsyncHTTPTransport):
                 connection = await self._get_connection_from_pool(origin)
 
                 if connection is None:
-                    connection = AsyncHTTPConnection(
-                        origin=origin,
-                        http2=self._http2,
-                        uds=self._uds,
-                        ssl_context=self._ssl_context,
-                        local_address=self._local_address,
-                    )
+                    connection = self._create_connection(origin=origin)
                     logger.trace("created connection=%r", connection)
                     await self._add_to_pool(connection, timeout=timeout)
                 else:
                     logger.trace("reuse connection=%r", connection)
 
             try:
-                response = await connection.request(
-                    method, url, headers=headers, stream=stream, timeout=timeout
+                response = await connection.arequest(
+                    method, url, headers=headers, stream=stream, ext=ext
                 )
             except NewConnectionRequired:
                 connection = None
-            except Exception:
+            except Exception:  # noqa: PIE786
                 logger.trace("remove from pool connection=%r", connection)
                 await self._remove_from_pool(connection)
                 raise
 
+        status_code, headers, stream, ext = response
         wrapped_stream = ResponseByteStream(
-            response[4], connection=connection, callback=self._response_closed
+            stream, connection=connection, callback=self._response_closed
         )
-        return response[0], response[1], response[2], response[3], wrapped_stream
+        return status_code, headers, wrapped_stream, ext
 
     async def _get_connection_from_pool(
         self, origin: Origin
@@ -265,7 +278,7 @@ class AsyncConnectionPool(AsyncHTTPTransport):
                 remove_from_pool = True
                 close_connection = True
             elif self._keepalive_expiry is not None:
-                now = self._backend.time()
+                now = await self._backend.time()
                 connection.expires_at = now + self._keepalive_expiry
 
         if remove_from_pool:
@@ -281,7 +294,7 @@ class AsyncConnectionPool(AsyncHTTPTransport):
         if self._keepalive_expiry is None:
             return
 
-        now = self._backend.time()
+        now = await self._backend.time()
         if now < self._next_keepalive_check:
             return
 
@@ -292,7 +305,7 @@ class AsyncConnectionPool(AsyncHTTPTransport):
             if (
                 connection.state == ConnectionState.IDLE
                 and connection.expires_at is not None
-                and now > connection.expires_at
+                and now >= connection.expires_at
             ):
                 connections_to_close.add(connection)
                 await self._remove_from_pool(connection)
@@ -301,10 +314,8 @@ class AsyncConnectionPool(AsyncHTTPTransport):
             await connection.aclose()
 
     async def _add_to_pool(
-        self, connection: AsyncHTTPConnection, timeout: TimeoutDict = None
+        self, connection: AsyncHTTPConnection, timeout: TimeoutDict
     ) -> None:
-        timeout = {} if timeout is None else timeout
-
         logger.trace("adding connection to pool=%r", connection)
         await self._connection_semaphore.acquire(timeout=timeout.get("pool", None))
         async with self._thread_lock:
@@ -315,7 +326,7 @@ class AsyncConnectionPool(AsyncHTTPTransport):
         logger.trace("removing connection from pool=%r", connection)
         async with self._thread_lock:
             if connection in self._connections.get(connection.origin, set()):
-                self._connection_semaphore.release()
+                await self._connection_semaphore.release()
                 self._connections[connection.origin].remove(connection)
                 if not self._connections[connection.origin]:
                     del self._connections[connection.origin]
@@ -346,7 +357,7 @@ class AsyncConnectionPool(AsyncHTTPTransport):
 
         stats = {}
         for origin, connections in self._connections.items():
-            stats[origin_to_url_string(origin)] = [
-                connection.info() for connection in connections
-            ]
+            stats[origin_to_url_string(origin)] = sorted(
+                [connection.info() for connection in connections]
+            )
         return stats
